import numpy as np

from loguru import logger


class optima:
    """
    stores all results and adaptations of it
    """

    def __init__(self, dim_x, max_optima):
        """
        input:
        -----
            dim ... the dimensionality of the space
            max_optima ... maximum number of stored optima
        """

        self.dim_x = dim_x
        self.max_optima = max_optima
        self.list = []

    ####################################################

    def make_optima_list_entry(self, x, f, classifier, eigs, grad, grad_norm, r):
        list_entry = {"x": x,
                      "f(x)": f,
                      "classifier": classifier,
                      "Hessian eigvals": eigs,
                      "df/dx": grad,
                      "|df/dx|": grad_norm,
                      "radius": r}

        return list_entry

    def fill_in_optima_list(self, res):
        x, f, g, eig, r, local_success = res[0], res[1], res[2], res[3], res[4], res[5]
        if not np.any(local_success) and len(self.list) == 0: local_success[:] = True
        clean_indices = np.where(np.asarray(local_success) == True)[0]
        if len(clean_indices) == 0: return self.list

        clean_x = x[clean_indices]
        clean_f = f[clean_indices]
        clean_g = g[clean_indices]
        clean_eig = eig[clean_indices]
        clean_radii = r[clean_indices]
        classifier = []
        ##making the classifier
        for i in range(len(clean_x)):
            if any(clean_g[i] > 1e-3):
                classifier.append("degenerate")
            elif any(abs(clean_eig[i]) < 10e-6):
                classifier.append("zero curvature")
            elif all(clean_eig[i] > 0.0):
                classifier.append("minimum")
            elif all(clean_eig[i] < 0.0):
                classifier.append("maximum")
            elif abs(np.max(np.sign(clean_eig[i])) - np.min(np.sign(clean_eig[i]))) == 2:
                classifier.append("saddle point")
            else:
                classifier.append("ERROR")

        new_optima_list = []
        for i in range(len(clean_x)):
            new_optima_list.append(
                self.make_optima_list_entry(clean_x[i], clean_f[i], classifier[i], clean_eig[i], clean_g[i],
                                            np.linalg.norm(clean_g[i]), clean_radii[i]))
        optima_list = self.list + new_optima_list

        def find_f(d): return d["f(x)"]

        optima_list.sort(key=find_f)
        self.list = optima_list[0:self.max_optima]
        return self.list

    def get_minima(self, n):
        try:
            minima_list = [entry for entry in self.list if entry["classifier"] == "minimum"]
            minima_list = minima_list[0:min(n, len(minima_list))]
            return minima_list
        except:
            logger.debug("no minima available in the optima_list")
            return None

    ####################################################
    def get_maxima(self, n):
        try:
            maxima_list = [entry for entry in self.list if entry["classifier"] == "maximum"]
            maxima_list = maxima_list[0:min(n, len(maxima_list))]
            return maxima_list
        except:
            logger.debug("no maxima available in the optima_list")
            return None

    ####################################################
    def get_deflation_points(self, n):
        try:
            defl_list = [entry for entry in self.list if
                         entry["classifier"] == "maximum" or entry["classifier"] == "minimum" or entry[
                             "classifier"] == "saddle point"]
            defl_x = [entry["x"] for entry in defl_list]
            defl_f = [entry["f(x)"] for entry in defl_list]
            defl_r = [entry["radius"] for entry in defl_list]

            return defl_x, defl_f, defl_r
        except Exception as e:
            logger.debug("no deflation points available in the optima_list")
            return [], [], []

#########################################################
#########################################################



from loguru import logger

try:
    from ._version import __version__
except (ImportError, ModuleNotFoundError) as ex:
    raise RuntimeError('Running hgdl from source code requires installation. If you would like an editable source '
                       'install, use "pip install -e ." to perform and editable installation.') from ex


logger.disable('hgdl')



import numpy as np
from loguru import logger
from .. import misc


def DNewton(func, grad, hess, bounds, x0, max_iter, tol, *args):
    e = np.inf
    gradient = np.ones((len(x0))) * np.inf
    counter = 0
    x = np.array(x0)
    grad_list = []
    while e > tol or np.max(abs(gradient)) > tol:
        x = misc.project_onto_bounds(x, bounds)
        x[abs(x) < 1e-16] = 0.
        gradient = grad(x, *args)
        gradient[abs(gradient) < 1e-16] = 0.
        hessian = hess(x, *args)
        # hessian = 0.5 * hessian @ hessian.T
        hessian[abs(hessian) < 1e-16] = 0.
        grad_list.append(np.max(gradient))
        try:
            gamma = np.linalg.solve(hessian, -gradient)
        except Exception as error:
            gamma, a, b, c = np.linalg.lstsq(hessian, -gradient, rcond=None)
        if any(gamma == np.nan) or any(gamma == np.inf): return x, func(x, *args), gradient, \
        np.linalg.eig(hess(x, *args))[0], False
        x += gamma
        e = np.max(abs(gamma))
        logger.debug("dNewton step size: ", e, " max gradient: ", np.max(abs(gradient)))
        if counter > max_iter: return x, func(x, *args), gradient, np.linalg.eig(hess(x, *args))[0], False
        counter += 1
    return x, func(x, *args), gradient, np.linalg.eig(hess(x, *args))[0], True



import time

import numpy as np


def out_of_bounds(x, bounds):
    for i in range(len(x)):
        if x[i] < bounds[i, 0] or x[i] > bounds[i, 1]:
            return True
    return False


def in_bounds(x, bounds):
    if (bounds[:, 1] - x > 0).all() and (bounds[:, 0] - x < 0).all():
        return True
    return False


def random_sample(N, k, bounds):
    sample = np.random.random((N, k))
    sample *= bounds[:, 1] - bounds[:, 0]
    sample += bounds[:, 0]
    return sample


def random_population(bounds, n):
    return np.random.uniform(low=bounds[:, 0], high=bounds[:, 1], size=(n, len(bounds)))


def finish_up_tasks(tasks):
    for f in tasks:
        if f.status == 'cancelled':
            tasks.remove(f)
    while any(f.status == 'pending' for f in tasks):
        time.sleep(0.1)
    return tasks


def project_onto_bounds(x, bounds):
    new_x = np.array(x)
    for i in range(len(x)):
        if x[i] < bounds[i, 0]: new_x[i] = bounds[i, 0]
        if x[i] > bounds[i, 1]: new_x[i] = bounds[i, 1]
    return new_x



import numpy as np
from distributed import get_client
from loguru import logger
from scipy.optimize import minimize

from . import bump_function as defl
from .. import misc
from .dNewton import DNewton as DNewton
import warnings


def run_local(d, optima, x0):
    x_defl, f_defl, radii = optima.get_deflation_points(len(optima.list))
    return run_local_optimizer(d, x0, x_defl, radii)


###########################################################################
def run_local_optimizer(d, x0, x_defl=[], radii=[]):
    """
    this function runs a deflated local methos for
    all the walkers.
    The loop below goes over every walker
    input:
        2d numpy array of initial positions
        2d numpy array of positions of deflations (optional, default = [])
    return:
        optima_locations, func values, gradient norms, eigenvalues, local_success(bool)
    """
    dim = d.dim
    number_of_walkers = d.number_of_walkers

    if len(x0) < number_of_walkers:
        x0 = np.row_stack([x0, misc.random_population(d.bounds, number_of_walkers - len(x0))])

    client = get_client()
    tasks = []
    bf = client.scatter(d, workers=d.workers["walkers"])
    for i in range(min(len(x0), number_of_walkers)):
        logger.debug("Worker ", i, " submitted")
        worker = d.workers["walkers"][(int(i - ((i // number_of_walkers) * number_of_walkers)))]
        data = {"d": bf, "x0": x0[i], "x_defl": x_defl, "radius": radii}
        tasks.append(client.submit(local_method, data, workers=worker))

    results = client.gather(tasks)
    number_of_walkers = len(tasks)
    x = np.empty((number_of_walkers, dim))
    f = np.empty((number_of_walkers))
    g = np.empty((number_of_walkers, dim))
    eig = np.empty((number_of_walkers, dim))
    r = np.empty((number_of_walkers))
    local_success = np.empty((number_of_walkers), dtype=bool)

    for i in range(len(tasks)):
        x[i], f[i], g[i], eig[i], r[i], local_success[i] = results[i]
        for j in range(i):
            if np.linalg.norm(np.subtract(x[i], x[j])) < r[i] and local_success[j] == True:
                logger.warning("points converged too close to each other in HGDL; point removed")
                local_success[i] = False
        for j in range(len(x_defl)):
            if np.linalg.norm(np.subtract(x[i], x_defl[j])) < radii[j] and all(g[i] < 1e-5):
                logger.warning("local method converged within 2 x radius of a deflated position in HGDL")
                local_success[i] = False
    return x, f, g, eig, r, local_success


def local_method(data, method="dNewton"):
    from functools import partial
    d = data["d"]
    x0 = np.array(data["x0"])
    e = np.inf
    local_success = False
    tol = d.tolerance
    x_defl = data["x_defl"]
    r_defl = data["radius"]
    bounds = d.bounds
    max_iter = d.local_max_iter
    args = d.args
    method = d.local_optimizer
    constr = d.constr
    # augment grad, hess
    grad = partial(defl.deflated_grad, grad_func=d.grad, x_defl=x_defl, radius=r_defl)
    hess = partial(defl.deflated_hess, grad_func=d.grad, hess_func=d.hess, x_defl=x_defl, radius=r_defl)

    # call local methods
    if method == "dNewton":
        x, f, g, eig, local_success = DNewton(d.func, grad, hess, bounds, x0, max_iter, tol, *args)
        if np.linalg.norm(g) < 1e-6 and np.min(eig) > 1e-6:
            local_success = True
            r = 1. / np.min(eig)
        else:
            eig = np.array([0.0])
            r = 0.0

    elif type(method) == str:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            res = minimize(d.func, x0, args=args, method=method, jac=grad, hess=hess,
            bounds=bounds, constraints=constr, tol = tol, options={"disp": False})
        x = res["x"]
        f = res["fun"]
        g = res["jac"]
        eig = np.linalg.eig(hess(x, *args))[0]

        if np.linalg.norm(g) < 1e-6 and np.min(eig) > 1e-6:
            local_success = True
            r = 1. / np.min(eig)
        else:
            eig = np.array([0.0])
            r = 0.0


    elif callable(method):
        res = method(d.func, grad, hess, bounds, x0, *args)
        x = res["x"]
        f = res["fun"]
        g = res["jac"]
        if np.linalg.norm(g) < 1e-6 and np.min(eig) > 1e-6:
            local_success = True
            eig = np.linalg.eig(hess(x, *args))[0]
            r = 1. / np.min(eig)
        else:
            eig = np.array([0.0])
            r = 0.0

    else:
        raise Exception("no local method specified")

    return x, f, g, np.real(eig), np.abs(r), local_success
###########################################################################



##global optimizer for hgdl
import numpy as np

from .. import misc


def run_global(x, y, bounds, method, number_of_offspring):
    # print("Global optimizer: ", method)
    if method == "genetic":
        return genetic_step(x, y, bounds, number_of_offspring)
    elif method == "random":
        return random_step(x, y, bounds, number_of_offspring)
    elif method is callable:
        return method(x, y, bounds, number_of_offspring)
    else:
        raise Exception("no global method specified")


def random_step(x, y, bounds, n):
    offspring = np.random.uniform(low=bounds[:, 0], high=bounds[:, 1], size=(n, len(bounds)))
    return offspring


###########################################################################
def genetic_step(X, y, bounds, numChoose):
    """
    Input:
    X is the individuals - points on a surface
    y is the performance - f(X)
    Notes:
    the children can be outside of the bounds!
    """
    unfairness = 100.
    wildness = 0.0005
    N, k = X.shape
    # normalize the performances to (0,1)
    y -= np.amin(y)
    amax = np.amax(y)
    # if the distribution of performance has no width,
    #   give everyone an equal shot
    if np.isclose(amax, 0.):
        p = np.ones(N) * 1. / N
    else:
        y /= np.amax(y)
        y *= -1.
        y -= np.amin(y)
        y += 1
        p = y / np.sum(y)
    # This chooses from the sample based on the power law,
    # allowing replacement means that the individuals
    # can have multiple kids
    p = unfairness * np.power(p, unfairness - 1)
    p /= np.sum(p)
    if np.isnan(p).any():
        raise Exception("got isnans in GeneticStep")
    moms = np.random.choice(N, size=numChoose, replace=True, p=p)
    dads = np.random.choice(N, size=numChoose, replace=True, p=p)
    # calculate a perturbation to the median
    #   of each individual's parents
    perturbation = np.random.normal(
        loc=0.,
        scale=wildness * (bounds[:, 1] - bounds[:, 0]),
        size=(numChoose, k))
    # the children are the median of their parents plus a perturbation
    norm = p[moms] + p[dads]
    weights = (p[moms] / norm, p[dads] / norm)
    weighted_linear_sum = weights[0].reshape(-1, 1) * X[moms] + weights[1].reshape(-1, 1) * X[dads]
    children = weighted_linear_sum + perturbation
    oob = np.logical_not([misc.in_bounds(x, bounds) for x in children])
    children[oob] = misc.random_sample(np.sum(oob), k, bounds)
    # print("=========================")
    # print("Children in HGDL genetic alg.:", flush = True)
    # print(children)
    # print("=========================")
    return children



###local optimizer for hgdl
import numpy as np


def deflated_grad(x, *args, grad_func=None, x_defl=[], radius=[]):
    d = deflation_function(x, x_defl, radius)
    return d * grad_func(x, *args)


def deflated_hess(x, *args, grad_func=None, hess_func=None, x_defl=[], radius=[]):
    d = deflation_function(x, x_defl, radius)
    dg = deflation_function_gradient(x, x_defl, radius)
    return (hess_func(x, *args) * d) + np.outer(grad_func(x, *args), dg)


########################################################
########################################################
########################################################
def b(x, x0, r):
    """
    evaluates the bump function
    x ... a point (1d numpy array)
    x0 ... 1d numpy array of location of bump function, a 2d numpy array

    returns the bump function b(x,x0) with radius r
    """
    d = np.sqrt((x - x0).T @ (x - x0))
    a = 1.0 - (d ** 2 / r ** 2)
    if a <= 0:
        return 0.0
    else:
        return np.exp(-1.0 / a) * np.exp(1.0)


###########################################################################
def b_grad(x, x0, r):
    """evaluates the bump function gradient b(x,x0) with radius r
       x ... point
       x0... location of bump
       r ... radius of bump function
    """
    d = np.sqrt((x - x0).T @ (x - x0))
    d2 = (x - x0)
    a = 1.0 - (d ** 2 / r ** 2)
    if a <= 0:
        return np.zeros((len(x)))
    else:
        return (b(x, x0, r) * ((-2.0 * d2) / (a ** 2))) / r ** 2


###########################################################################
def deflation_function(x, x0, r):
    """
    input:
        x is one point(1d numpy array)
        x0 is a a 2d array of locations of the bump function
    the return is the deflation operator, e.g. 1.0/(1.0 - bump(x,x0))
    """
    if len(x0) == 0: return 1.0
    s = 0.0
    for i in range(len(x0)):
        s += b(x, x0[i], r[i])
    if s == 1.0: s = 0.99999
    return 1.0 / (1.0 - s)


###########################################################################
def deflation_function_gradient(x, x0, r):
    """
    input:
        x is one point(1d numpy array)
        x0 is a 2d array of locations of the bump function
    the return is the gradient of the deflation operator, e.g. (1.0/(1.0 - bump(x,x0)))'
    """
    if len(x0) == 0: return np.zeros((len(x)))
    s1 = 0.0
    s2 = np.zeros((len(x)))
    for i in range(len(x0)):
        s1 += b(x, x0[i], r[i])
        s2 += b_grad(x, x0[i], r[i])
    if s1 == 1.0: s1 = 0.99999
    return s2 / ((1.0 - s1) ** 2)



import warnings

import dask.distributed as distributed
import dask.multiprocessing
import numpy as np
from loguru import logger

from . import misc
from .global_methods.global_optimizer import run_global
from .local_methods.local_optimizer import run_local
from .meta_data import meta_data
from .optima import optima


class HGDL:
    """
    This is HGDL, a class for asynchronous HPC-capable optimization. \n
    H ... Hybrid \n
    G ... Global \n
    D ... Deflated \n
    L ... Local \n
    The algorithm places a number of walkers inside the domain 
    (the number is determined by the dask client), all of which perform
    a local optimization in a distributed way in parallel. 
    When the walkers have identified local optima, their 
    positions are communicated back to the host
    who removes the found optima by deflation, 
    and replaces the fittest walkers by a global 
    optimization step. From here the next epoch
    begins with distributed local optimizations of 
    the new walkers. The algorithm results in a 
    sorted list of unique optima (only if optima
    are of the form f'(x) = 0)
    The method `hgdl.optimize` instantly 
    returns a result object that can be 
    queried for a growing, sorted list of
    optima. If a Hessian is provided, 
    those optima are classified as minima, 
    maxima or saddle points.

    Parameters
    ----------
    func : Callable
        The function to be MINIMIZED. A callable that accepts an np.ndarray and 
        optional arguments, and returns a scalar.
    grad : Callable
        The gradient of the function to be MINIMIZED. A callable that accepts an
        np.ndarray and optional arguments, and returns a vector
        (np.ndarray) of shape (D), where D is the dimensionality of the space in
        which the
        optimization takes place.
    bounds : np.ndarray
        The bounds of the domain; an np.ndarray of shape (D x 2), where D is the
        dimensionality of the space in which the
        optimization takes place. Here D is the dimension of the input domain.
    hess : Callable, optional
        The Hessian of the function to be MINIMIZED. A callable that accepts an 
        np.ndarray and optional arguments, and returns a
        np.ndarray of shape (D x D). The default value is no-op.
    num_epochs : int, optional
        The number of epochs the algorithm runs through before being terminated.
        One epoch is the convergence of all local walkers,
        the deflation of the identified optima, and the global replacement of the 
        walkers. Note, the algorithm is running asynchronously, so a high number
        of epochs can be chosen without concerns, it will not affect the run time 
        to obtain the optima. Therefore, the default is
        100000.
    global_optimizer : Callable or str, optional
        The function (identified by a string or a Callable) that replaces the 
        fittest walkers after their local convergence.
        The possible options are `genetic` (default), `random` or a callable that 
        accepts an
        np.ndarray of shape (U x D) of positions, an np.ndarray of shape (U) of 
        function values,
        and np.ndarray of shape (D x 2) of bounds, and an integer specifying the
        number of offspring
        individuals that should be returned. The callable should return the 
        positions of the offspring
        individuals as an np.ndarray of shape (number_of_offspring x D).
    local_optimizer : Callable or str, optional
        The local optimizer that is used. The options are
        `dNewton` (default), `L-BFGS-B`, `BFGS`, `CG`, `Newton-CG`, `SLSQP`.
        The above methods have been tested, but most others should work. Visit 
        the `scipy.optimize.minimize` docs 
        (https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html)
        for specifications and limitations of the local methods. The parameter 
        also accepts a callable of the form func(f,grad,hess,bounds,x0,*args), 
        and returns an object equal to the scipy.optimize.minimize methods.
    number_of_optima : int, optional
        The number of optima that will be stored in the optima list and deflated.
        The default is 1e6.
        After that number is reached, worse-performing optima will not be stored 
        or deflated.
    local_max_iter : int, optional
        The number of iterations before local optimizations are terminated. 
        The default is 1000.
        It can be lowered when second-order local optimizers are used.
    args : tuple, optional
        A tuple of arguments that will be communicated to the function, 
        the gradient, and the Hessian callables.
        Default = ().
    constraints : object, optional
        An optional n-tuple of constraint objects.
        The default is no constraints (). Constraints are defined following 
        scipy.optimize.NonlinearConstraint.

    Attributes
    ----------
    optima : object
        Contains the attribute optima.list in which the optima are stored.
        However, the method 'get_latest()' should be used to access the optima.


    """

    def __init__(self, func, grad, bounds,
                 hess=None, num_epochs=100000,
                 global_optimizer="genetic",
                 local_optimizer="L-BFGS-B",
                 number_of_optima=1000000,
                 local_max_iter=1000,
                 constraints=(),
                 args=()):
        bounds = np.asarray(bounds)
        self.dim = len(bounds)
        self.bounds = bounds
        self.func = func
        self.grad = grad
        if hess:
            self.hess = hess
        else:
            self.hess = self.hess_approx
        if bounds is not None and local_optimizer == "dNewton":
            warnings.warn("Warning: dNewton will not adhere to bounds. It is recommended to formulate your objective function such that it is defined on R^N by simple non-linear transformations.")
        if constraints:
            local_optimizer = "SLSQP"
            warnings.warn("Constraints provided, local optimizer changed to 'SLSQP'")

        self.constraints = constraints
        self.local_max_iter = local_max_iter
        self.num_epochs = num_epochs
        self.global_optimizer = global_optimizer
        self.local_optimizer = local_optimizer
        self.args = args
        self.optima = optima(self.dim, number_of_optima)
        logger.debug("HGDL successfully initiated {}")
        if callable(self.hess): logger.debug("Hessian was provided by the user: {}", self.hess)
        logger.debug("========================")

    ###########################################################################
    ###########################################################################
    ############USER FUNCTIONS#################################################
    ###########################################################################
    ###########################################################################
    ###########################################################################
    def optimize(self, dask_client=None, x0=None, tolerance=1e-10):
        """
        Function to start the optimization. Note, this function will not 
        return anything. Use the method hgdl.HGDL.get_latest() 
        (non-blocking) or hgdl.HGDL.get_final() (blocking)
        to query results.

        Parameters
        ----------
        dask_client : distributed.client.Client, optional
            The client that will be used for the distibuted local 
            optimizations. The default is a local client.
        x0 : np.ndarray, optional
            An np.ndarray of shape (V x D) of points used as 
            starting positions. If V > number of walkers 
            (specified by the dask client) the array will be truncated.
            If V < number of walkers, random points will be appended.
            The default is None, meaning only random points will be used.
        tolerance : float, optional
            The tolerance used by the local optimizers. The default is 1e-6
        """
        client = self._init_dask_client(dask_client)
        self.tolerance = tolerance
        logger.debug(client)
        self.x0 = self._prepare_starting_positions(x0)
        logger.debug("HGDL starts with: {}", self.x0)
        self.meta_data = meta_data(self)
        self._run_epochs(client)

    ###########################################################################
    def get_client_info(self):
        """
        Function to receive info about the workers.
        """
        return self.workers

    ###########################################################################
    def get_latest(self):
        """
        Function to request the current result.
        No inputs
        """
        try:
            data, frames = self.transfer_data.get()
            self.optima = distributed.protocol.deserialize(data, frames)
            logger.debug("HGDL called get_latest() successfully")
        except Exception as err:
            self.optima = self.optima
            logger.error("HGDL get_latest failed due to {} \n optima list unchanged", str(err))
        optima_list = self.optima.list
        return optima_list

    ###########################################################################
    def get_final(self):
        """
        Function to request the final result.
        CAUTION: This function will block the main thread until
        the end of all epochs is reached.
        No inputs.
        """
        try:
            self.optima = self.main_future.result()
        except Exception as err:
            logger.error("HGDL get_final failed due to {}", str(err))
        optima_list = self.optima.list
        return optima_list

    ###########################################################################
    def cancel_tasks(self):
        """
        Function to cancel all tasks and therefore the execution.
        However, this function does not kill the client.
        """
        logger.debug("HGDL is cancelling all tasks...")
        res = self.get_latest()
        self.break_condition.set(True)
        self.client.cancel(self.main_future)
        logger.debug("Status of HGDL task: ", self.main_future.status)
        logger.debug("This leaves the client alive.")
        return res

    ###########################################################################
    def kill_client(self):
        """
        Function to cancel all tasks and kill the dask client, 
        and therefore the execution.
        If cancel_tasks() is called before, this will throw an error.
        """
        logger.debug("HGDL kill client initialized ...")
        res = self.get_latest()
        try:
            self.break_condition.set(True)
            self.client.cancel(self.main_future)
            self.client.close()
            logger.debug("HGDL kill client successful")
        except Exception as err:
            raise RuntimeError("HGDL kill failed") from err
        return res

    ###########################################################################
    ############USER FUNCTIONS END#############################################
    ###########################################################################
    def _prepare_starting_positions(self, x0):
        if x0 is not None and len(x0[0]) != self.dim:
            raise Exception("Wrong dimensionality of starting positions")
        elif x0 is None:
            x0 = misc.random_population(self.bounds, self.number_of_walkers)
        elif x0.ndim == 1:
            x0 = np.array([x0])

        if len(x0) < self.number_of_walkers:
            x0_aux = np.zeros((self.number_of_walkers, len(x0[0])))
            x0_aux[0:len(x0)] = x0
            x0_aux[len(x0):] = misc.random_population(self.bounds, self.number_of_walkers - len(x0))
            x0 = x0_aux
        elif len(x0) > self.number_of_walkers:
            x0 = x0[0:self.number_of_walkers]
        else:
            x0 = x0
        return x0

    ###########################################################################
    def _init_dask_client(self, dask_client):
        if dask_client is None:
            dask_client = dask.distributed.Client()
            logger.debug("No dask client provided to HGDL. Using the local client")
        else:
            logger.debug("dask client provided to HGDL")
        client = dask_client
        worker_info = list(client.scheduler_info()["workers"].keys())
        if not worker_info: raise Exception("No workers available")
        self.workers = {"host": worker_info[0],
                        "walkers": worker_info[1:]}
        logger.debug(f"Host {self.workers['host']} has {len(self.workers['walkers'])} workers.")
        self.number_of_walkers = len(self.workers["walkers"])
        return client

    ###########################################################################
    def _run_epochs(self, client):
        self.break_condition = distributed.Variable("break_condition", client)
        self.transfer_data = distributed.Variable("transfer_data", client)
        a = distributed.protocol.serialize(self.optima)
        self.transfer_data.set(a)
        self.break_condition.set(False)
        data = {"transfer data": self.transfer_data,
                "break condition": self.break_condition,
                "optima": self.optima, "metadata": self.meta_data}
        bf = client.scatter(data, workers=self.workers["host"])
        self.main_future = client.submit(hgdl, bf, workers=self.workers["host"])
        self.client = client

    ###########################################################################
    def hess_approx(self, x, *args):
        ##implements a first-order approximation
        len_x = len(x)
        hess = np.zeros((len_x, len_x))
        epsilon = 1e-6
        grad_x = self.grad(x, *args)
        for i in range(len_x):
            x_temp = np.array(x)
            x_temp[i] = x_temp[i] + epsilon
            hess[i, i:] = ((self.grad(x_temp, *args) - grad_x) / epsilon)[i:]
        return hess + hess.T - np.diag(np.diag(hess))


###########################################################################
###########################################################################
##################hgdl functions###########################################
###########################################################################
###########################################################################
def hgdl(data):
    metadata = data["metadata"]
    transfer_data = data["transfer data"]
    break_condition = data["break condition"]
    optima = data["optima"]
    logger.debug("HGDL computing epoch 1 of {}", metadata.num_epochs)
    res = run_local(metadata,optima,metadata.x0)
    logger.debug("filling in optima list for the first time.", flush = True)
    optima.fill_in_optima_list(res)
    logger.debug("optima list filled", flush = True)
    a = distributed.protocol.serialize(optima)
    transfer_data.set(a)

    logger.debug("HGDL first local optimization round done.", flush = True)
    for i in range(1, metadata.num_epochs):
        bc = break_condition.get()
        if bc is True:
            logger.debug(f"HGDL Epoch {i} was cancelled")
            break
        logger.debug(f"HGDL computing epoch {i + 1} of ", metadata.num_epochs)
        optima = run_hgdl_epoch(metadata, optima)
        a = distributed.protocol.serialize(optima)
        transfer_data.set(a)
    logger.debug("HGDL finished all epochs!")
    return optima


###########################################################################
def run_hgdl_epoch(metadata, optima):
    optima_list = optima.list
    n = min(len(optima_list),metadata.number_of_walkers)
    ind_pos = [entry["x"] for entry in optima_list]
    ind_fit = [entry["f(x)"] for entry in optima_list]

    global_res = run_global(\
            np.array(ind_pos[:n]),
            np.array(ind_fit[0:n]),
            metadata.bounds[0:metadata.dim], metadata.global_optimizer,n)
    x0 = np.zeros((n,metadata.dim))
    x0[:,0:metadata.dim] = np.array(global_res)
    res = run_local(metadata,optima,x0)
    op = optima.fill_in_optima_list(res)
    return optima
